﻿namespace BitBadger.Documents.Postgres

/// <summary>The type of index to generate for the document</summary>
[<Struct>]
type DocumentIndex =

    /// <summary>A <tt>GIN</tt> index with standard operations (all operators supported)</summary>
    | Full

    /// <summary>
    /// A <tt>GIN</tt> index with JSONPath operations (optimized for <tt>@&gt;</tt>, <tt>@?</tt>, <tt>@@</tt> operators)
    /// </summary>
    | Optimized


open Npgsql

/// <summary>Configuration for document handling</summary>
module Configuration =

    /// The data source to use for query execution
    let mutable private dataSourceValue : NpgsqlDataSource option = None

    /// <summary>Register a data source to use for query execution (disposes the current one if it exists)</summary>
    /// <param name="source">The data source to use</param>
    [<CompiledName "UseDataSource">]
    let useDataSource source =
        if Option.isSome dataSourceValue then dataSourceValue.Value.Dispose()
        dataSourceValue <- Some source

    /// <summary>Retrieve the currently configured data source</summary>
    /// <returns>The current data source</returns>
    /// <exception cref="T:System.InvalidOperationException">If no data source has been configured</exception>
    [<CompiledName "DataSource">]
    let dataSource () =
        match dataSourceValue with
        | Some source -> source
        | None -> invalidOp "Please provide a data source before attempting data access"


open Npgsql.FSharp

/// Helper functions
[<AutoOpen>]
module private Helpers =

    /// Shorthand to retrieve the data source as SqlProps
    let internal fromDataSource () =
        Configuration.dataSource () |> Sql.fromDataSource

    /// Execute a task and ignore the result
    let internal ignoreTask<'T> (it : System.Threading.Tasks.Task<'T>) = backgroundTask {
        let! _ = it
        ()
    }

    /// Create a number or string parameter, or use the given parameter derivation function if non-(numeric or string)
    let internal parameterFor<'T> (value: 'T) (catchAllFunc: 'T -> SqlValue) =
        match box value with
        | :? int8    as it -> Sql.int8 it
        | :? uint8   as it -> Sql.int8 (int8 it)
        | :? int16   as it -> Sql.int16 it
        | :? uint16  as it -> Sql.int16 (int16 it)
        | :? int     as it -> Sql.int it
        | :? uint32  as it -> Sql.int (int it)
        | :? int64   as it -> Sql.int64 it
        | :? uint64  as it -> Sql.int64 (int64 it)
        | :? decimal as it -> Sql.decimal it
        | :? single  as it -> Sql.double (double it)
        | :? double  as it -> Sql.double it
        | :? string  as it -> Sql.string it
        | _                -> catchAllFunc value


open BitBadger.Documents

/// <summary>Functions for creating parameters</summary>
[<AutoOpen>]
module Parameters =

    /// <summary>Create an ID parameter (name "@id")</summary>
    /// <param name="key">The key value for the ID parameter</param>
    /// <returns>The name and parameter value for the ID</returns>
    [<CompiledName "Id">]
    let idParam (key: 'TKey) =
        "@id", parameterFor key (fun it -> Sql.string (string it))

    /// <summary>Create a parameter with a JSON value</summary>
    /// <param name="name">The name of the parameter to create</param>
    /// <param name="it">The criteria to provide as JSON</param>
    /// <returns>The name and parameter value for the JSON field</returns>
    [<CompiledName "Json">]
    let jsonParam (name: string) (it: 'TJson) =
        name, Sql.jsonb (Configuration.serializer().Serialize it)

    /// <summary>Create JSON field parameters</summary>
    /// <param name="fields">The <tt>Field</tt>s to convert to parameters</param>
    /// <param name="parameters">The current parameters for the query</param>
    /// <returns>A unified sequence of parameter names and values</returns>
    [<CompiledName "AddFields">]
    let addFieldParams fields parameters =
        let name = ParameterName()
        fields
        |> Seq.map (fun it ->
            seq {
                match it.Comparison with
                | Exists | NotExists -> ()
                | Between (min, max) ->
                    let p = name.Derive it.ParameterName
                    yield ($"{p}min", parameterFor min (fun v -> Sql.parameter (NpgsqlParameter($"{p}min", v))))
                    yield ($"{p}max", parameterFor max (fun v -> Sql.parameter (NpgsqlParameter($"{p}max", v))))
                | In values ->
                    let p = name.Derive it.ParameterName
                    yield!
                        values
                        |> Seq.mapi (fun idx v ->
                            let paramName = $"{p}_{idx}"
                            paramName, Sql.parameter (NpgsqlParameter(paramName, v)))
                | InArray (_, values) ->
                    let p = name.Derive it.ParameterName
                    yield (p, Sql.stringArray (values |> Seq.map string |> Array.ofSeq))
                | Equal v | Greater v | GreaterOrEqual v | Less v | LessOrEqual v | NotEqual v ->
                    let p = name.Derive it.ParameterName
                    yield (p, parameterFor v (fun l -> Sql.parameter (NpgsqlParameter(p, l)))) })
        |> Seq.collect id
        |> Seq.append parameters
        |> Seq.toList
        |> Seq.ofList

    /// <summary>Append JSON field name parameters for the given field names to the given parameters</summary>
    /// <param name="fieldNames">The names of fields to be addressed</param>
    /// <returns>The name (<tt>@name</tt>) and parameter value for the field names</returns>
    [<CompiledName "FieldNames">]
    let fieldNameParams (fieldNames: string seq) =
        if Seq.length fieldNames = 1 then "@name", Sql.string (Seq.head fieldNames)
        else "@name", Sql.stringArray (Array.ofSeq fieldNames)

    /// <summary>An empty parameter sequence</summary>
    [<CompiledName "None">]
    let noParams =
        Seq.empty<string * SqlValue>


/// <summary>Query construction functions</summary>
[<RequireQualifiedAccess>]
module Query =

    /// <summary>
    /// Create a <tt>WHERE</tt> clause fragment to implement a comparison on fields in a JSON document
    /// </summary>
    /// <param name="howMatched">How the fields should be matched</param>
    /// <param name="fields">The fields for the comparisons</param>
    /// <returns>A <tt>WHERE</tt> clause implementing the comparisons for the given fields</returns>
    [<CompiledName "WhereByFields">]
    let whereByFields (howMatched: FieldMatch) fields =
        let name = ParameterName()
        let isNumeric (it: obj) =
            match it with
            | :? int8    | :? uint8  | :? int16  | :? uint16 | :? int | :? uint32 | :? int64 | :? uint64
            | :? decimal | :? single | :? double -> true
            | _ -> false
        fields
        |> Seq.map (fun it ->
            match it.Comparison with
            | Exists | NotExists -> $"{it.Path PostgreSQL AsSql} {it.Comparison.OpSql}"
            | InArray _ -> $"{it.Path PostgreSQL AsJson} {it.Comparison.OpSql} {name.Derive it.ParameterName}"
            | _ ->
                let p            = name.Derive it.ParameterName
                let param, value =
                    match it.Comparison with
                    | Between (min, _) -> $"{p}min AND {p}max", min
                    | In values ->
                        let paramNames = values |> Seq.mapi (fun idx _ -> $"{p}_{idx}") |> String.concat ", "
                        $"({paramNames})", defaultArg (Seq.tryHead values) (obj ())
                    | Equal v | Greater v | GreaterOrEqual v | Less v | LessOrEqual v | NotEqual v  -> p, v
                    | _ -> p, ""
                if isNumeric value then
                    $"({it.Path PostgreSQL AsSql})::numeric {it.Comparison.OpSql} {param}"
                else $"{it.Path PostgreSQL AsSql} {it.Comparison.OpSql} {param}")
        |> String.concat $" {howMatched} "

    /// <summary>Create a <tt>WHERE</tt> clause fragment to implement an ID-based query</summary>
    /// <param name="docId">The ID of the document</param>
    /// <returns>A <tt>WHERE</tt> clause fragment identifying a document by its ID</returns>
    [<CompiledName "WhereById">]
    let whereById<'TKey> (docId: 'TKey) =
        whereByFields Any [ { Field.Equal (Configuration.idField ()) docId with ParameterName = Some "@id" } ]

    /// <summary>Table and index definition queries</summary>
    module Definition =

        /// <summary>SQL statement to create a document table</summary>
        /// <param name="name">The name of the table (may include schema)</param>
        /// <returns>A query to create the table if it does not exist</returns>
        [<CompiledName "EnsureTable">]
        let ensureTable name =
            Query.Definition.ensureTableFor name "JSONB"

        /// <summary>SQL statement to create an index on JSON documents in the specified table</summary>
        /// <param name="name">The name of the table to be indexed (may include schema)</param>
        /// <param name="idxType">The type of document index to create</param>
        /// <returns>A query to create the index if it does not exist</returns>
        [<CompiledName "EnsureDocumentIndex">]
        let ensureDocumentIndex (name: string) idxType =
            let extraOps = match idxType with Full -> "" | Optimized -> " jsonb_path_ops"
            let tableName = name.Split '.' |> Array.last
            $"CREATE INDEX IF NOT EXISTS idx_{tableName}_document ON {name} USING GIN (data{extraOps})"

    /// <summary>
    /// Create a <tt>WHERE</tt> clause fragment to implement a <tt>@&gt;</tt> (JSON contains) condition
    /// </summary>
    /// <param name="paramName">The parameter name for the query</param>
    /// <returns>A <tt>WHERE</tt> clause fragment for the contains condition</returns>
    [<CompiledName "WhereDataContains">]
    let whereDataContains paramName =
        $"data @> %s{paramName}"

    /// <summary>
    /// Create a <tt>WHERE</tt> clause fragment to implement a <tt>@?</tt> (JSON Path match) condition
    /// </summary>
    /// <param name="paramName">The parameter name for the query</param>
    /// <returns>A <tt>WHERE</tt> clause fragment for the JSON Path match condition</returns>
    [<CompiledName "WhereJsonPathMatches">]
    let whereJsonPathMatches paramName =
        $"data @? %s{paramName}::jsonpath"

    /// <summary>Create an <tt>UPDATE</tt> statement to patch documents</summary>
    /// <param name="tableName">The table to be updated</param>
    /// <returns>A query to patch documents</returns>
    [<CompiledName "Patch">]
    let patch tableName =
        $"UPDATE %s{tableName} SET data = data || @data"

    /// <summary>Create an <tt>UPDATE</tt> statement to remove fields from documents</summary>
    /// <param name="tableName">The table to be updated</param>
    /// <returns>A query to remove fields from documents</returns>
    [<CompiledName "RemoveFields">]
    let removeFields tableName =
        $"UPDATE %s{tableName} SET data = data - @name"

    /// <summary>Create a query by a document's ID</summary>
    /// <param name="statement">The SQL statement to be run against a document by its ID</param>
    /// <param name="docId">The ID of the document targeted</param>
    /// <returns>A query addressing a document by its ID</returns>
    [<CompiledName "ById">]
    let byId<'TKey> statement (docId: 'TKey) =
        Query.statementWhere statement (whereById docId)

    /// <summary>Create a query on JSON fields</summary>
    /// <param name="statement">The SQL statement to be run against matching fields</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to be matched</param>
    /// <returns>A query addressing documents by field matching conditions</returns>
    [<CompiledName "ByFields">]
    let byFields statement howMatched fields =
        Query.statementWhere statement (whereByFields howMatched fields)

    /// <summary>Create a JSON containment query</summary>
    /// <param name="statement">The SQL statement to be run against the containment query</param>
    /// <returns>A query addressing documents by a JSON containment query</returns>
    [<CompiledName "ByContains">]
    let byContains statement =
        Query.statementWhere statement (whereDataContains "@criteria")

    /// <summary>Create a JSON Path match query</summary>
    /// <param name="statement">The SQL statement to run against the JSON Path match</param>
    /// <returns>A query addressing documents by a JSON Path match</returns>
    [<CompiledName "ByPathMatch">]
    let byPathMatch statement =
        Query.statementWhere statement (whereJsonPathMatches "@path")


/// <summary>Functions for dealing with results</summary>
[<AutoOpen>]
module Results =

    /// <summary>Create a domain item from a document, specifying the field in which the document is found</summary>
    /// <param name="field">The field name containing the JSON document</param>
    /// <param name="row">A row reader set to the row with the document to be constructed</param>
    /// <returns>The constructed domain item</returns>
    [<CompiledName "FromDocument">]
    let fromDocument<'T> field (row: RowReader) : 'T =
        Configuration.serializer().Deserialize<'T>(row.string field)

    /// <summary>Create a domain item from a document</summary>
    /// <param name="row">A row reader set to the row with the document to be constructed</param>
    /// <returns>The constructed domain item</returns>
    [<CompiledName "FromData">]
    let fromData<'T> row : 'T =
        fromDocument "data" row

    /// <summary>Extract a count from the column <tt>it</tt></summary>
    /// <param name="row">A row reader set to the row with the count to retrieve</param>
    /// <returns>The count from the row</returns>
    [<CompiledName "ToCount">]
    let toCount (row: RowReader) =
        row.int "it"

    /// <summary>Extract a true/false value from the column <tt>it</tt></summary>
    /// <param name="row">A row reader set to the row with the true/false value to retrieve</param>
    /// <returns>The true/false value from the row</returns>
    [<CompiledName "ToExists">]
    let toExists (row: RowReader) =
        row.bool "it"


/// <summary>Versions of queries that accept <tt>SqlProps</tt> as the last parameter</summary>
module WithProps =

    module FSharpList = Microsoft.FSharp.Collections.List

    /// <summary>Commands to execute custom SQL queries</summary>
    [<RequireQualifiedAccess>]
    module Custom =

        /// <summary>Execute a query that returns a list of results</summary>
        /// <param name="query">The query to retrieve the results</param>
        /// <param name="parameters">Parameters to use for the query</param>
        /// <param name="mapFunc">The mapping function between the document and the domain item</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>A list of results for the given query</returns>
        [<CompiledName "FSharpList">]
        let list<'TDoc> query parameters (mapFunc: RowReader -> 'TDoc) sqlProps =
            Sql.query query sqlProps
            |> Sql.parameters (List.ofSeq parameters)
            |> Sql.executeAsync mapFunc

        /// <summary>Execute a query that returns a list of results</summary>
        /// <param name="query">The query to retrieve the results</param>
        /// <param name="parameters">Parameters to use for the query</param>
        /// <param name="mapFunc">The mapping function between the document and the domain item</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>A list of results for the given query</returns>
        let List<'TDoc>(query, parameters, mapFunc: System.Func<RowReader, 'TDoc>, sqlProps) = backgroundTask {
            let! results = list<'TDoc> query parameters mapFunc.Invoke sqlProps
            return ResizeArray results
        }

        /// <summary>Execute a query that returns one or no results</summary>
        /// <param name="query">The query to retrieve the results</param>
        /// <param name="parameters">Parameters to use for the query</param>
        /// <param name="mapFunc">The mapping function between the document and the domain item</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns><tt>Some</tt> with the first matching result, or <tt>None</tt> if not found</returns>
        [<CompiledName "FSharpSingle">]
        let single<'TDoc> query parameters mapFunc sqlProps = backgroundTask {
            let! results = list<'TDoc> query parameters mapFunc sqlProps
            return FSharpList.tryHead results
        }

        /// <summary>Execute a query that returns one or no results</summary>
        /// <param name="query">The query to retrieve the results</param>
        /// <param name="parameters">Parameters to use for the query</param>
        /// <param name="mapFunc">The mapping function between the document and the domain item</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>The first matching result, or <tt>null</tt> if not found</returns>
        let Single<'TDoc when 'TDoc: null and 'TDoc: not struct>(
                query, parameters, mapFunc: System.Func<RowReader, 'TDoc>, sqlProps) = backgroundTask {
            let! result = single<'TDoc> query parameters mapFunc.Invoke sqlProps
            return Option.toObj result
        }

        /// <summary>Execute a query that returns no results</summary>
        /// <param name="query">The query to retrieve the results</param>
        /// <param name="parameters">Parameters to use for the query</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "NonQuery">]
        let nonQuery query parameters sqlProps =
            Sql.query query sqlProps
            |> Sql.parameters (FSharpList.ofSeq parameters)
            |> Sql.executeNonQueryAsync
            |> ignoreTask

        /// <summary>Execute a query that returns a scalar value</summary>
        /// <param name="query">The query to retrieve the value</param>
        /// <param name="parameters">Parameters to use for the query</param>
        /// <param name="mapFunc">The mapping function to obtain the value</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>The scalar value for the query</returns>
        [<CompiledName "FSharpScalar">]
        let scalar<'T when 'T: struct> query parameters (mapFunc: RowReader -> 'T) sqlProps =
            Sql.query query sqlProps
            |> Sql.parameters (FSharpList.ofSeq parameters)
            |> Sql.executeRowAsync mapFunc

        /// <summary>Execute a query that returns a scalar value</summary>
        /// <param name="query">The query to retrieve the value</param>
        /// <param name="parameters">Parameters to use for the query</param>
        /// <param name="mapFunc">The mapping function to obtain the value</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>The scalar value for the query</returns>
        let Scalar<'T when 'T: struct>(query, parameters, mapFunc: System.Func<RowReader, 'T>, sqlProps) =
            scalar<'T> query parameters mapFunc.Invoke sqlProps

    /// <summary>Table and index definition commands</summary>
    module Definition =

        /// <summary>Create a document table</summary>
        /// <param name="name">The table whose existence should be ensured (may include schema)</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "EnsureTable">]
        let ensureTable name sqlProps = backgroundTask {
            do! Custom.nonQuery (Query.Definition.ensureTable name) [] sqlProps
            do! Custom.nonQuery (Query.Definition.ensureKey name PostgreSQL) [] sqlProps
        }

        /// <summary>Create an index on documents in the specified table</summary>
        /// <param name="name">The table to be indexed (may include schema)</param>
        /// <param name="idxType">The type of document index to create</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "EnsureDocumentIndex">]
        let ensureDocumentIndex name idxType sqlProps =
            Custom.nonQuery (Query.Definition.ensureDocumentIndex name idxType) [] sqlProps

        /// <summary>Create an index on field(s) within documents in the specified table</summary>
        /// <param name="tableName">The table to be indexed (may include schema)</param>
        /// <param name="indexName">The name of the index to create</param>
        /// <param name="fields">One or more fields to be indexed</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "EnsureFieldIndex">]
        let ensureFieldIndex tableName indexName fields sqlProps =
            Custom.nonQuery (Query.Definition.ensureIndexOn tableName indexName fields PostgreSQL) [] sqlProps

    /// <summary>Commands to add documents</summary>
    [<AutoOpen>]
    module Document =

        /// <summary>Insert a new document</summary>
        /// <param name="tableName">The table into which the document should be inserted (may include schema)</param>
        /// <param name="document">The document to be inserted</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "Insert">]
        let insert<'TDoc> tableName (document: 'TDoc) sqlProps =
            let query =
                match Configuration.autoIdStrategy () with
                | Disabled -> Query.insert tableName
                | strategy ->
                    let idField   = Configuration.idField ()
                    let dataParam =
                        if AutoId.NeedsAutoId strategy document idField then
                            match strategy with
                            | Number ->
                                $"' || (SELECT COALESCE(MAX((data->>'{idField}')::numeric), 0) + 1 FROM {tableName}) || '"
                            | Guid -> $"\"{AutoId.GenerateGuid()}\""
                            | RandomString -> $"\"{AutoId.GenerateRandomString(Configuration.idStringLength ())}\""
                            | Disabled -> "@data"
                            |> function it -> $"""@data::jsonb || ('{{"{idField}":{it}}}')::jsonb"""
                        else "@data"
                    (Query.insert tableName).Replace("@data", dataParam)
            Custom.nonQuery query [ jsonParam "@data" document ] sqlProps

        /// <summary>
        /// Save a document, inserting it if it does not exist and updating it if it does (AKA "upsert")
        /// </summary>
        /// <param name="tableName">The table into which the document should be saved (may include schema)</param>
        /// <param name="document">The document to be saved</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "Save">]
        let save<'TDoc> tableName (document: 'TDoc) sqlProps =
            Custom.nonQuery (Query.save tableName) [ jsonParam "@data" document ] sqlProps

    /// <summary>Commands to count documents</summary>
    [<RequireQualifiedAccess>]
    module Count =

        /// <summary>Count all documents in a table</summary>
        /// <param name="tableName">The table in which documents should be counted (may include schema)</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>The count of the documents in the table</returns>
        [<CompiledName "All">]
        let all tableName sqlProps =
            Custom.scalar (Query.count tableName) [] toCount sqlProps

        /// <summary>Count matching documents using JSON field comparisons (<tt>-&gt;&gt; =</tt>, etc.)</summary>
        /// <param name="tableName">The table in which documents should be counted (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>The count of matching documents in the table</returns>
        [<CompiledName "ByFields">]
        let byFields tableName howMatched fields sqlProps =
            Custom.scalar
                (Query.byFields (Query.count tableName) howMatched fields) (addFieldParams fields []) toCount sqlProps

        /// <summary>Count matching documents using a JSON containment query (<tt>@&gt;</tt>)</summary>
        /// <param name="tableName">The table in which documents should be counted (may include schema)</param>
        /// <param name="criteria">The document to match with the containment query</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>The count of the documents in the table</returns>
        [<CompiledName "ByContains">]
        let byContains tableName (criteria: 'TContains) sqlProps =
            Custom.scalar
                (Query.byContains (Query.count tableName)) [ jsonParam "@criteria" criteria ] toCount sqlProps

        /// <summary>Count matching documents using a JSON Path match query (<tt>@?</tt>)</summary>
        /// <param name="tableName">The table in which documents should be counted (may include schema)</param>
        /// <param name="jsonPath">The JSON Path expression to be matched</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>The count of the documents in the table</returns>
        [<CompiledName "ByJsonPath">]
        let byJsonPath tableName jsonPath sqlProps =
            Custom.scalar
                (Query.byPathMatch (Query.count tableName)) [ "@path", Sql.string jsonPath ] toCount sqlProps

    /// <summary>Commands to determine if documents exist</summary>
    [<RequireQualifiedAccess>]
    module Exists =

        /// <summary>Determine if a document exists for the given ID</summary>
        /// <param name="tableName">The table in which existence should be checked (may include schema)</param>
        /// <param name="docId">The ID of the document whose existence should be checked</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>True if a document exists, false if not</returns>
        [<CompiledName "ById">]
        let byId tableName (docId: 'TKey) sqlProps =
            Custom.scalar (Query.exists tableName (Query.whereById docId)) [ idParam docId ] toExists sqlProps

        /// <summary>Determine if a document exists using JSON field comparisons (<tt>-&gt;&gt; =</tt>, etc.)</summary>
        /// <param name="tableName">The table in which existence should be checked (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>True if any matching documents exist, false if not</returns>
        [<CompiledName "ByFields">]
        let byFields tableName howMatched fields sqlProps =
            Custom.scalar
                (Query.exists tableName (Query.whereByFields howMatched fields))
                (addFieldParams fields [])
                toExists
                sqlProps

        /// <summary>Determine if a document exists using a JSON containment query (<tt>@&gt;</tt>)</summary>
        /// <param name="tableName">The table in which existence should be checked (may include schema)</param>
        /// <param name="criteria">The document to match with the containment query</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>True if any matching documents exist, false if not</returns>
        [<CompiledName "ByContains">]
        let byContains tableName (criteria: 'TContains) sqlProps =
            Custom.scalar
                (Query.exists tableName (Query.whereDataContains "@criteria"))
                [ jsonParam "@criteria" criteria ]
                toExists
                sqlProps

        /// <summary>Determine if a document exists using a JSON Path match query (<tt>@?</tt>)</summary>
        /// <param name="tableName">The table in which existence should be checked (may include schema)</param>
        /// <param name="jsonPath">The JSON Path expression to be matched</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>True if any matching documents exist, false if not</returns>
        [<CompiledName "ByJsonPath">]
        let byJsonPath tableName jsonPath sqlProps =
            Custom.scalar
                (Query.exists tableName (Query.whereJsonPathMatches "@path"))
                [ "@path", Sql.string jsonPath ]
                toExists
                sqlProps

    /// <summary>Commands to determine if documents exist</summary>
    [<RequireQualifiedAccess>]
    module Find =

        /// <summary>Retrieve all documents in the given table</summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>All documents from the given table</returns>
        [<CompiledName "FSharpAll">]
        let all<'TDoc> tableName sqlProps =
            Custom.list<'TDoc> (Query.find tableName) [] fromData<'TDoc> sqlProps

        /// <summary>Retrieve all documents in the given table</summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>All documents from the given table</returns>
        let All<'TDoc>(tableName, sqlProps) =
            Custom.List<'TDoc>(Query.find tableName, [], fromData<'TDoc>, sqlProps)

        /// <summary>Retrieve all documents in the given table ordered by the given fields in the document</summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>All documents from the given table, ordered by the given fields</returns>
        [<CompiledName "FSharpAllOrdered">]
        let allOrdered<'TDoc> tableName orderFields sqlProps =
            Custom.list<'TDoc> (Query.find tableName + Query.orderBy orderFields PostgreSQL) [] fromData<'TDoc> sqlProps

        /// <summary>Retrieve all documents in the given table ordered by the given fields in the document</summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>All documents from the given table, ordered by the given fields</returns>
        let AllOrdered<'TDoc>(tableName, orderFields, sqlProps) =
            Custom.List<'TDoc>(
                Query.find tableName + Query.orderBy orderFields PostgreSQL, [], fromData<'TDoc>, sqlProps)

        /// <summary>Retrieve a document by its ID</summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="docId">The ID of the document to retrieve</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns><tt>Some</tt> with the document if found, <tt>None</tt> otherwise</returns>
        [<CompiledName "FSharpById">]
        let byId<'TKey, 'TDoc> tableName (docId: 'TKey) sqlProps =
            Custom.single (Query.byId (Query.find tableName) docId) [ idParam docId ] fromData<'TDoc> sqlProps

        /// <summary>Retrieve a document by its ID</summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="docId">The ID of the document to retrieve</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>The document if found, <tt>null</tt> otherwise</returns>
        let ById<'TKey, 'TDoc when 'TDoc: null and 'TDoc: not struct>(tableName, docId: 'TKey, sqlProps) =
            Custom.Single<'TDoc>(
                Query.byId (Query.find tableName) docId, [ idParam docId ], fromData<'TDoc>, sqlProps)

        /// <summary>Retrieve documents matching JSON field comparisons (<tt>-&gt;&gt; =</tt>, etc.)</summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>All documents matching the given fields</returns>
        [<CompiledName "FSharpByFields">]
        let byFields<'TDoc> tableName howMatched fields sqlProps =
            Custom.list<'TDoc>
                (Query.byFields (Query.find tableName) howMatched fields)
                (addFieldParams fields [])
                fromData<'TDoc>
                sqlProps

        /// <summary>Retrieve documents matching JSON field comparisons (<tt>-&gt;&gt; =</tt>, etc.)</summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>All documents matching the given fields</returns>
        let ByFields<'TDoc>(tableName, howMatched, fields, sqlProps) =
            Custom.List<'TDoc>(
                Query.byFields (Query.find tableName) howMatched fields,
                addFieldParams fields [],
                fromData<'TDoc>,
                sqlProps)

        /// <summary>
        /// Retrieve documents matching JSON field comparisons (<tt>-&gt;&gt; =</tt>, etc.) ordered by the given fields
        /// in the document
        /// </summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="queryFields">The field conditions to match</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>All documents matching the given fields, ordered by the other given fields</returns>
        [<CompiledName "FSharpByFieldsOrdered">]
        let byFieldsOrdered<'TDoc> tableName howMatched queryFields orderFields sqlProps =
            Custom.list<'TDoc>
                (Query.byFields (Query.find tableName) howMatched queryFields + Query.orderBy orderFields PostgreSQL)
                (addFieldParams queryFields [])
                fromData<'TDoc>
                sqlProps

        /// <summary>
        /// Retrieve documents matching JSON field comparisons (<tt>-&gt;&gt; =</tt>, etc.) ordered by the given fields
        /// in the document
        /// </summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="queryFields">The field conditions to match</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>All documents matching the given fields, ordered by the other given fields</returns>
        let ByFieldsOrdered<'TDoc>(tableName, howMatched, queryFields, orderFields, sqlProps) =
            Custom.List<'TDoc>(
                Query.byFields (Query.find tableName) howMatched queryFields + Query.orderBy orderFields PostgreSQL,
                addFieldParams queryFields [],
                fromData<'TDoc>,
                sqlProps)

        /// <summary>Retrieve documents matching a JSON containment query (<tt>@&gt;</tt>)</summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="criteria">The document to match with the containment query</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>All documents matching the given containment query</returns>
        [<CompiledName "FSharpByContains">]
        let byContains<'TDoc> tableName (criteria: obj) sqlProps =
            Custom.list<'TDoc>
                (Query.byContains (Query.find tableName)) [ jsonParam "@criteria" criteria ] fromData<'TDoc> sqlProps

        /// <summary>Retrieve documents matching a JSON containment query (<tt>@&gt;</tt>)</summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="criteria">The document to match with the containment query</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>All documents matching the given containment query</returns>
        let ByContains<'TDoc>(tableName, criteria: obj, sqlProps) =
            Custom.List<'TDoc>(
                Query.byContains (Query.find tableName),
                [ jsonParam "@criteria" criteria ],
                fromData<'TDoc>,
                sqlProps)

        /// <summary>
        /// Retrieve documents matching a JSON containment query (<tt>@&gt;</tt>) ordered by the given fields in the
        /// document
        /// </summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="criteria">The document to match with the containment query</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>All documents matching the given containment query, ordered by the given fields</returns>
        [<CompiledName "FSharpByContainsOrdered">]
        let byContainsOrdered<'TDoc> tableName (criteria: obj) orderFields sqlProps =
            Custom.list<'TDoc>
                (Query.byContains (Query.find tableName) + Query.orderBy orderFields PostgreSQL)
                [ jsonParam "@criteria" criteria ]
                fromData<'TDoc>
                sqlProps

        /// <summary>
        /// Retrieve documents matching a JSON containment query (<tt>@&gt;</tt>) ordered by the given fields in the
        /// document
        /// </summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="criteria">The document to match with the containment query</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>All documents matching the given containment query, ordered by the given fields</returns>
        let ByContainsOrdered<'TDoc>(tableName, criteria: obj, orderFields, sqlProps) =
            Custom.List<'TDoc>(
                Query.byContains (Query.find tableName) + Query.orderBy orderFields PostgreSQL,
                [ jsonParam "@criteria" criteria ],
                fromData<'TDoc>,
                sqlProps)

        /// <summary>Retrieve documents matching a JSON Path match query (<tt>@?</tt>)</summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="jsonPath">The JSON Path expression to match</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>All documents matching the given JSON Path expression</returns>
        [<CompiledName "FSharpByJsonPath">]
        let byJsonPath<'TDoc> tableName jsonPath sqlProps =
            Custom.list<'TDoc>
                (Query.byPathMatch (Query.find tableName)) [ "@path", Sql.string jsonPath ] fromData<'TDoc> sqlProps

        /// <summary>Retrieve documents matching a JSON Path match query (<tt>@?</tt>)</summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="jsonPath">The JSON Path expression to match</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>All documents matching the given JSON Path expression</returns>
        let ByJsonPath<'TDoc>(tableName, jsonPath, sqlProps) =
            Custom.List<'TDoc>(
                Query.byPathMatch (Query.find tableName),
                [ "@path", Sql.string jsonPath ],
                fromData<'TDoc>,
                sqlProps)

        /// <summary>
        /// Retrieve documents matching a JSON Path match query (<tt>@?</tt>) ordered by the given fields in the
        /// document
        /// </summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="jsonPath">The JSON Path expression to match</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>All documents matching the given JSON Path expression, ordered by the given fields</returns>
        [<CompiledName "FSharpByJsonPathOrdered">]
        let byJsonPathOrdered<'TDoc> tableName jsonPath orderFields sqlProps =
            Custom.list<'TDoc>
                (Query.byPathMatch (Query.find tableName) + Query.orderBy orderFields PostgreSQL)
                [ "@path", Sql.string jsonPath ]
                fromData<'TDoc>
                sqlProps

        /// <summary>
        /// Retrieve documents matching a JSON Path match query (<tt>@?</tt>) ordered by the given fields in the
        /// document
        /// </summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="jsonPath">The JSON Path expression to match</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>All documents matching the given JSON Path expression, ordered by the given fields</returns>
        let ByJsonPathOrdered<'TDoc>(tableName, jsonPath, orderFields, sqlProps) =
            Custom.List<'TDoc>(
                Query.byPathMatch (Query.find tableName) + Query.orderBy orderFields PostgreSQL,
                [ "@path", Sql.string jsonPath ],
                fromData<'TDoc>,
                sqlProps)

        /// <summary>Retrieve the first document matching JSON field comparisons (<tt>-&gt;&gt; =</tt>, etc.)</summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns><tt>Some</tt> with the first document, or <tt>None</tt> if not found</returns>
        [<CompiledName "FSharpFirstByFields">]
        let firstByFields<'TDoc> tableName howMatched fields sqlProps =
            Custom.single<'TDoc>
                $"{Query.byFields (Query.find tableName) howMatched fields} LIMIT 1"
                (addFieldParams fields [])
                fromData<'TDoc>
                sqlProps

        /// <summary>Retrieve the first document matching JSON field comparisons (<tt>-&gt;&gt; =</tt>, etc.)</summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>The first document, or <tt>null</tt> if not found</returns>
        let FirstByFields<'TDoc when 'TDoc: null and 'TDoc: not struct>(tableName, howMatched, fields, sqlProps) =
            Custom.Single<'TDoc>(
                $"{Query.byFields (Query.find tableName) howMatched fields} LIMIT 1",
                addFieldParams fields [],
                fromData<'TDoc>,
                sqlProps)

        /// <summary>
        /// Retrieve the first document matching JSON field comparisons (<tt>-&gt;&gt; =</tt>, etc.) ordered by the
        /// given fields in the document
        /// </summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="queryFields">The field conditions to match</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>
        /// <tt>Some</tt> with the first document ordered by the given fields, or <tt>None</tt> if not found
        /// </returns>
        [<CompiledName "FSharpFirstByFieldsOrdered">]
        let firstByFieldsOrdered<'TDoc> tableName howMatched queryFields orderFields sqlProps =
            Custom.single<'TDoc>
                $"{Query.byFields (Query.find tableName) howMatched queryFields}{Query.orderBy orderFields PostgreSQL} LIMIT 1"
                (addFieldParams queryFields [])
                fromData<'TDoc>
                sqlProps

        /// <summary>
        /// Retrieve the first document matching JSON field comparisons (<tt>-&gt;&gt; =</tt>, etc.) ordered by the
        /// given fields in the document
        /// </summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="queryFields">The field conditions to match</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>The first document ordered by the given fields, or <tt>null</tt> if not found</returns>
        let FirstByFieldsOrdered<'TDoc when 'TDoc: null and 'TDoc: not struct>(
                tableName, howMatched, queryFields, orderFields, sqlProps) =
            Custom.Single<'TDoc>(
                $"{Query.byFields (Query.find tableName) howMatched queryFields}{Query.orderBy orderFields PostgreSQL} LIMIT 1",
                addFieldParams queryFields [],
                fromData<'TDoc>,
                sqlProps)

        /// <summary>Retrieve the first document matching a JSON containment query (<tt>@&gt;</tt>)</summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="criteria">The document to match with the containment query</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns><tt>Some</tt> with the first document, or <tt>None</tt> if not found</returns>
        [<CompiledName "FSharpFirstByContains">]
        let firstByContains<'TDoc> tableName (criteria: obj) sqlProps =
            Custom.single<'TDoc>
                $"{Query.byContains (Query.find tableName)} LIMIT 1"
                [ jsonParam "@criteria" criteria ]
                fromData<'TDoc>
                sqlProps

        /// <summary>Retrieve the first document matching a JSON containment query (<tt>@&gt;</tt>)</summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="criteria">The document to match with the containment query</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>The first document, or <tt>null</tt> if not found</returns>
        let FirstByContains<'TDoc when 'TDoc: null and 'TDoc: not struct>(tableName, criteria: obj, sqlProps) =
            Custom.Single<'TDoc>(
                $"{Query.byContains (Query.find tableName)} LIMIT 1",
                [ jsonParam "@criteria" criteria ],
                fromData<'TDoc>,
                sqlProps)

        /// <summary>
        /// Retrieve the first document matching a JSON containment query (<tt>@&gt;</tt>) ordered by the given fields
        /// in the document
        /// </summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="criteria">The document to match with the containment query</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>
        /// <tt>Some</tt> with the first document ordered by the given fields, or <tt>None</tt> if not found
        /// </returns>
        [<CompiledName "FSharpFirstByContainsOrdered">]
        let firstByContainsOrdered<'TDoc> tableName (criteria: obj) orderFields sqlProps =
            Custom.single<'TDoc>
                $"{Query.byContains (Query.find tableName)}{Query.orderBy orderFields PostgreSQL} LIMIT 1"
                [ jsonParam "@criteria" criteria ]
                fromData<'TDoc>
                sqlProps

        /// <summary>
        /// Retrieve the first document matching a JSON containment query (<tt>@&gt;</tt>) ordered by the given fields
        /// in the document
        /// </summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="criteria">The document to match with the containment query</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>The first document ordered by the given fields, or <tt>null</tt> if not found</returns>
        let FirstByContainsOrdered<'TDoc when 'TDoc: null and 'TDoc: not struct>(
                tableName, criteria: obj, orderFields, sqlProps) =
            Custom.Single<'TDoc>(
                $"{Query.byContains (Query.find tableName)}{Query.orderBy orderFields PostgreSQL} LIMIT 1",
                [ jsonParam "@criteria" criteria ],
                fromData<'TDoc>,
                sqlProps)

        /// <summary>Retrieve the first document matching a JSON Path match query (<tt>@?</tt>)</summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="jsonPath">The JSON Path expression to match</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns><tt>Some</tt> with the first document, or <tt>None</tt> if not found</returns>
        [<CompiledName "FSharpFirstByJsonPath">]
        let firstByJsonPath<'TDoc> tableName jsonPath sqlProps =
            Custom.single<'TDoc>
                $"{Query.byPathMatch (Query.find tableName)} LIMIT 1"
                [ "@path", Sql.string jsonPath ]
                fromData<'TDoc>
                sqlProps

        /// <summary>Retrieve the first document matching a JSON Path match query (<tt>@?</tt>)</summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="jsonPath">The JSON Path expression to match</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>The first document, or <tt>null</tt> if not found</returns>
        let FirstByJsonPath<'TDoc when 'TDoc: null and 'TDoc: not struct>(tableName, jsonPath, sqlProps) =
            Custom.Single<'TDoc>(
                $"{Query.byPathMatch (Query.find tableName)} LIMIT 1",
                [ "@path", Sql.string jsonPath ],
                fromData<'TDoc>,
                sqlProps)

        /// <summary>
        /// Retrieve the first document matching a JSON Path match query (<tt>@?</tt>) ordered by the given fields in
        /// the document
        /// </summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="jsonPath">The JSON Path expression to match</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>
        /// <tt>Some</tt> with the first document ordered by the given fields, or <tt>None</tt> if not found
        /// </returns>
        [<CompiledName "FSharpFirstByJsonPathOrdered">]
        let firstByJsonPathOrdered<'TDoc> tableName jsonPath orderFields sqlProps =
            Custom.single<'TDoc>
                $"{Query.byPathMatch (Query.find tableName)}{Query.orderBy orderFields PostgreSQL} LIMIT 1"
                [ "@path", Sql.string jsonPath ]
                fromData<'TDoc>
                sqlProps

        /// <summary>
        /// Retrieve the first document matching a JSON Path match query (<tt>@?</tt>) ordered by the given fields in
        /// the document
        /// </summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="jsonPath">The JSON Path expression to match</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        /// <returns>The first document ordered by the given fields, or <tt>null</tt> if not found</returns>
        let FirstByJsonPathOrdered<'TDoc when 'TDoc: null and 'TDoc: not struct>(
                tableName, jsonPath, orderFields, sqlProps) =
            Custom.Single<'TDoc>(
                $"{Query.byPathMatch (Query.find tableName)}{Query.orderBy orderFields PostgreSQL} LIMIT 1",
                [ "@path", Sql.string jsonPath ],
                fromData<'TDoc>,
                sqlProps)

    /// <summary>Commands to update documents</summary>
    [<RequireQualifiedAccess>]
    module Update =

        /// <summary>Update (replace) an entire document by its ID</summary>
        /// <param name="tableName">The table in which a document should be updated (may include schema)</param>
        /// <param name="docId">The ID of the document to be updated (replaced)</param>
        /// <param name="document">The new document</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "ById">]
        let byId tableName (docId: 'TKey) (document: 'TDoc) sqlProps =
            Custom.nonQuery
                (Query.byId (Query.update tableName) docId) [ idParam docId; jsonParam "@data" document ] sqlProps

        /// <summary>
        /// Update (replace) an entire document by its ID, using the provided function to obtain the ID from the
        /// document
        /// </summary>
        /// <param name="tableName">The table in which a document should be updated (may include schema)</param>
        /// <param name="idFunc">The function to obtain the ID of the document</param>
        /// <param name="document">The new document</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "FSharpByFunc">]
        let byFunc tableName (idFunc: 'TDoc -> 'TKey) (document: 'TDoc) sqlProps =
            byId tableName (idFunc document) document sqlProps

        /// <summary>
        /// Update (replace) an entire document by its ID, using the provided function to obtain the ID from the
        /// document
        /// </summary>
        /// <param name="tableName">The table in which a document should be updated (may include schema)</param>
        /// <param name="idFunc">The function to obtain the ID of the document</param>
        /// <param name="document">The new document</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        let ByFunc(tableName, idFunc: System.Func<'TDoc, 'TKey>, document: 'TDoc, sqlProps) =
            byFunc tableName idFunc.Invoke document sqlProps

    /// <summary>Commands to patch (partially update) documents</summary>
    [<RequireQualifiedAccess>]
    module Patch =

        /// <summary>Patch a document by its ID</summary>
        /// <param name="tableName">The table in which a document should be patched (may include schema)</param>
        /// <param name="docId">The ID of the document to patch</param>
        /// <param name="patch">The partial document to patch the existing document</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "ById">]
        let byId tableName (docId: 'TKey) (patch: 'TPatch) sqlProps =
            Custom.nonQuery
                (Query.byId (Query.patch tableName) docId) [ idParam docId; jsonParam "@data" patch ] sqlProps

        /// <summary>
        /// Patch documents using a JSON field comparison query in the <tt>WHERE</tt> clause (<tt>-&gt;&gt; =</tt>,
        /// etc.)
        /// </summary>
        /// <param name="tableName">The table in which documents should be patched (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        /// <param name="patch">The partial document to patch the existing document</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "ByFields">]
        let byFields tableName howMatched fields (patch: 'TPatch) sqlProps =
            Custom.nonQuery
                (Query.byFields (Query.patch tableName) howMatched fields)
                (addFieldParams fields [ jsonParam "@data" patch ])
                sqlProps

        /// <summary>Patch documents using a JSON containment query in the WHERE clause (<tt>@&gt;</tt>)</summary>
        /// <param name="tableName">The table in which documents should be patched (may include schema)</param>
        /// <param name="criteria">The document to match the containment query</param>
        /// <param name="patch">The partial document to patch the existing document</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "ByContains">]
        let byContains tableName (criteria: 'TContains) (patch: 'TPatch) sqlProps =
            Custom.nonQuery
                (Query.byContains (Query.patch tableName))
                [ jsonParam "@data" patch; jsonParam "@criteria" criteria ]
                sqlProps

        /// <summary>Patch documents using a JSON Path match query in the WHERE clause (<tt>@?</tt>)</summary>
        /// <param name="tableName">The table in which documents should be patched (may include schema)</param>
        /// <param name="jsonPath">The JSON Path expression to match</param>
        /// <param name="patch">The partial document to patch the existing document</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "ByJsonPath">]
        let byJsonPath tableName jsonPath (patch: 'TPatch) sqlProps =
            Custom.nonQuery
                (Query.byPathMatch (Query.patch tableName))
                [ jsonParam "@data" patch; "@path", Sql.string jsonPath ]
                sqlProps

    /// <summary>Commands to remove fields from documents</summary>
    [<RequireQualifiedAccess>]
    module RemoveFields =

        /// <summary>Remove fields from a document by the document's ID</summary>
        /// <param name="tableName">The table in which a document should be modified (may include schema)</param>
        /// <param name="docId">The ID of the document to modify</param>
        /// <param name="fieldNames">One or more field names to remove from the document</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "ById">]
        let byId tableName (docId: 'TKey) fieldNames sqlProps =
            Custom.nonQuery
                (Query.byId (Query.removeFields tableName) docId) [ idParam docId; fieldNameParams fieldNames ] sqlProps

        /// <summary>Remove fields from documents via a comparison on JSON fields in the document</summary>
        /// <param name="tableName">The table in which documents should be modified (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        /// <param name="fieldNames">One or more field names to remove from the matching documents</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "ByFields">]
        let byFields tableName howMatched fields fieldNames sqlProps =
            Custom.nonQuery
                (Query.byFields (Query.removeFields tableName) howMatched fields)
                (addFieldParams fields [ fieldNameParams fieldNames ])
                sqlProps

        /// <summary>Remove fields from documents via a JSON containment query (<tt>@&gt;</tt>)</summary>
        /// <param name="tableName">The table in which documents should be modified (may include schema)</param>
        /// <param name="criteria">The document to match the containment query</param>
        /// <param name="fieldNames">One or more field names to remove from the matching documents</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "ByContains">]
        let byContains tableName (criteria: 'TContains) fieldNames sqlProps =
            Custom.nonQuery
                (Query.byContains (Query.removeFields tableName))
                [ jsonParam "@criteria" criteria; fieldNameParams fieldNames ]
                sqlProps

        /// <summary>Remove fields from documents via a JSON Path match query (<tt>@?</tt>)</summary>
        /// <param name="tableName">The table in which documents should be modified (may include schema)</param>
        /// <param name="jsonPath">The JSON Path expression to match</param>
        /// <param name="fieldNames">One or more field names to remove from the matching documents</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "FSharpByJsonPath">]
        let byJsonPath tableName jsonPath fieldNames sqlProps =
            Custom.nonQuery
                (Query.byPathMatch (Query.removeFields tableName))
                [ "@path", Sql.string jsonPath; fieldNameParams fieldNames ]
                sqlProps

    /// <summary>Commands to delete documents</summary>
    [<RequireQualifiedAccess>]
    module Delete =

        /// <summary>Delete a document by its ID</summary>
        /// <param name="tableName">The table in which a document should be deleted (may include schema)</param>
        /// <param name="docId">The ID of the document to delete</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "ById">]
        let byId tableName (docId: 'TKey) sqlProps =
            Custom.nonQuery (Query.byId (Query.delete tableName) docId) [ idParam docId ] sqlProps

        /// <summary>Delete documents by matching a JSON field comparison query (<tt>-&gt;&gt; =</tt>, etc.)</summary>
        /// <param name="tableName">The table in which documents should be deleted (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "ByFields">]
        let byFields tableName howMatched fields sqlProps =
            Custom.nonQuery
                (Query.byFields (Query.delete tableName) howMatched fields) (addFieldParams fields []) sqlProps

        /// <summary>Delete documents by matching a JSON contains query (<tt>@&gt;</tt>)</summary>
        /// <param name="tableName">The table in which documents should be deleted (may include schema)</param>
        /// <param name="criteria">The document to match the containment query</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "ByContains">]
        let byContains tableName (criteria: 'TCriteria) sqlProps =
            Custom.nonQuery (Query.byContains (Query.delete tableName)) [ jsonParam "@criteria" criteria ] sqlProps

        /// <summary>Delete documents by matching a JSON Path match query (@?)</summary>
        /// <param name="tableName">The table in which documents should be deleted (may include schema)</param>
        /// <param name="jsonPath">The JSON Path expression to match</param>
        /// <param name="sqlProps">The <tt>SqlProps</tt> to use to execute the query</param>
        [<CompiledName "ByJsonPath">]
        let byJsonPath tableName jsonPath sqlProps =
            Custom.nonQuery (Query.byPathMatch (Query.delete tableName)) [ "@path", Sql.string jsonPath ] sqlProps


/// Commands to execute custom SQL queries
[<RequireQualifiedAccess>]
module Custom =

    /// Execute a query that returns a list of results
    [<CompiledName "FSharpList">]
    let list<'TDoc> query parameters (mapFunc: RowReader -> 'TDoc) =
        WithProps.Custom.list<'TDoc> query parameters mapFunc (fromDataSource ())

    /// Execute a query that returns a list of results
    let List<'TDoc>(query, parameters, mapFunc: System.Func<RowReader, 'TDoc>) =
        WithProps.Custom.List<'TDoc>(query, parameters, mapFunc, fromDataSource ())

    /// Execute a query that returns one or no results; returns None if not found
    [<CompiledName "FSharpSingle">]
    let single<'TDoc> query parameters (mapFunc: RowReader ->  'TDoc) =
        WithProps.Custom.single<'TDoc> query parameters mapFunc (fromDataSource ())

    /// Execute a query that returns one or no results; returns null if not found
    let Single<'TDoc when 'TDoc: null and 'TDoc: not struct>(
            query, parameters, mapFunc: System.Func<RowReader, 'TDoc>) =
        WithProps.Custom.Single<'TDoc>(query, parameters, mapFunc, fromDataSource ())

    /// Execute a query that returns no results
    [<CompiledName "NonQuery">]
    let nonQuery query parameters =
        WithProps.Custom.nonQuery query parameters (fromDataSource ())

    /// Execute a query that returns a scalar value
    [<CompiledName "FSharpScalar">]
    let scalar<'T when 'T: struct> query parameters (mapFunc: RowReader -> 'T) =
        WithProps.Custom.scalar query parameters mapFunc (fromDataSource ())

    /// Execute a query that returns a scalar value
    let Scalar<'T when 'T: struct>(query, parameters, mapFunc: System.Func<RowReader, 'T>) =
        WithProps.Custom.Scalar<'T>(query, parameters, mapFunc, fromDataSource ())


/// Table and index definition commands
[<RequireQualifiedAccess>]
module Definition =

    /// Create a document table
    [<CompiledName "EnsureTable">]
    let ensureTable name =
        WithProps.Definition.ensureTable name (fromDataSource ())

    /// Create an index on documents in the specified table
    [<CompiledName "EnsureDocumentIndex">]
    let ensureDocumentIndex name idxType =
        WithProps.Definition.ensureDocumentIndex name idxType (fromDataSource ())

    /// Create an index on field(s) within documents in the specified table
    [<CompiledName "EnsureFieldIndex">]
    let ensureFieldIndex tableName indexName fields =
        WithProps.Definition.ensureFieldIndex tableName indexName fields (fromDataSource ())


/// Document writing functions
[<AutoOpen>]
module Document =

    /// Insert a new document
    [<CompiledName "Insert">]
    let insert<'TDoc> tableName (document: 'TDoc) =
        WithProps.Document.insert<'TDoc> tableName document (fromDataSource ())

    /// Save a document, inserting it if it does not exist and updating it if it does (AKA "upsert")
    [<CompiledName "Save">]
    let save<'TDoc> tableName (document: 'TDoc) =
        WithProps.Document.save<'TDoc> tableName document (fromDataSource ())


/// Queries to count documents
[<RequireQualifiedAccess>]
module Count =

    /// Count all documents in a table
    [<CompiledName "All">]
    let all tableName =
        WithProps.Count.all tableName (fromDataSource ())

    /// Count matching documents using a JSON field comparison query (->> =)
    [<CompiledName "ByFields">]
    let byFields tableName howMatched fields =
        WithProps.Count.byFields tableName howMatched fields (fromDataSource ())

    /// Count matching documents using a JSON containment query (@>)
    [<CompiledName "ByContains">]
    let byContains tableName criteria =
        WithProps.Count.byContains tableName criteria (fromDataSource ())

    /// Count matching documents using a JSON Path match query (@?)
    [<CompiledName "ByJsonPath">]
    let byJsonPath tableName jsonPath =
        WithProps.Count.byJsonPath tableName jsonPath (fromDataSource ())


/// Queries to determine if documents exist
[<RequireQualifiedAccess>]
module Exists =

    /// Determine if a document exists for the given ID
    [<CompiledName "ById">]
    let byId tableName docId =
        WithProps.Exists.byId tableName docId (fromDataSource ())

    /// Determine if documents exist using a JSON field comparison query (->> =)
    [<CompiledName "ByFields">]
    let byFields tableName howMatched fields =
        WithProps.Exists.byFields tableName howMatched fields (fromDataSource ())

    /// Determine if documents exist using a JSON containment query (@>)
    [<CompiledName "ByContains">]
    let byContains tableName criteria =
        WithProps.Exists.byContains tableName criteria (fromDataSource ())

    /// Determine if documents exist using a JSON Path match query (@?)
    [<CompiledName "ByJsonPath">]
    let byJsonPath tableName jsonPath =
        WithProps.Exists.byJsonPath tableName jsonPath (fromDataSource ())


/// Commands to retrieve documents
[<RequireQualifiedAccess>]
module Find =

    /// Retrieve all documents in the given table
    [<CompiledName "FSharpAll">]
    let all<'TDoc> tableName =
        WithProps.Find.all<'TDoc> tableName (fromDataSource ())

    /// Retrieve all documents in the given table
    let All<'TDoc> tableName =
        WithProps.Find.All<'TDoc>(tableName, fromDataSource ())

    /// Retrieve all documents in the given table ordered by the given fields in the document
    [<CompiledName "FSharpAllOrdered">]
    let allOrdered<'TDoc> tableName orderFields =
        WithProps.Find.allOrdered<'TDoc> tableName orderFields (fromDataSource ())

    /// Retrieve all documents in the given table ordered by the given fields in the document
    let AllOrdered<'TDoc> tableName orderFields =
        WithProps.Find.AllOrdered<'TDoc>(tableName, orderFields, fromDataSource ())

    /// Retrieve a document by its ID; returns None if not found
    [<CompiledName "FSharpById">]
    let byId<'TKey, 'TDoc> tableName docId =
        WithProps.Find.byId<'TKey, 'TDoc> tableName docId (fromDataSource ())

    /// Retrieve a document by its ID; returns null if not found
    let ById<'TKey, 'TDoc when 'TDoc: null and 'TDoc: not struct>(tableName, docId: 'TKey) =
        WithProps.Find.ById<'TKey, 'TDoc>(tableName, docId, fromDataSource ())

    /// Retrieve documents matching a JSON field comparison query (->> =)
    [<CompiledName "FSharpByFields">]
    let byFields<'TDoc> tableName howMatched fields =
        WithProps.Find.byFields<'TDoc> tableName howMatched fields (fromDataSource ())

    /// Retrieve documents matching a JSON field comparison query (->> =)
    let ByFields<'TDoc>(tableName, howMatched, fields) =
        WithProps.Find.ByFields<'TDoc>(tableName, howMatched, fields, fromDataSource ())

    /// Retrieve documents matching a JSON field comparison query (->> =) ordered by the given fields in the document
    [<CompiledName "FSharpByFieldsOrdered">]
    let byFieldsOrdered<'TDoc> tableName howMatched queryField orderFields =
        WithProps.Find.byFieldsOrdered<'TDoc> tableName howMatched queryField orderFields (fromDataSource ())

    /// Retrieve documents matching a JSON field comparison query (->> =) ordered by the given fields in the document
    let ByFieldsOrdered<'TDoc>(tableName, howMatched, queryFields, orderFields) =
        WithProps.Find.ByFieldsOrdered<'TDoc>(tableName, howMatched, queryFields, orderFields, fromDataSource ())

    /// Retrieve documents matching a JSON containment query (@>)
    [<CompiledName "FSharpByContains">]
    let byContains<'TDoc> tableName (criteria: obj) =
        WithProps.Find.byContains<'TDoc> tableName criteria (fromDataSource ())

    /// Retrieve documents matching a JSON containment query (@>)
    let ByContains<'TDoc>(tableName, criteria: obj) =
        WithProps.Find.ByContains<'TDoc>(tableName, criteria, fromDataSource ())

    /// Retrieve documents matching a JSON containment query (@>) ordered by the given fields in the document
    [<CompiledName "FSharpByContainsOrdered">]
    let byContainsOrdered<'TDoc> tableName (criteria: obj) orderFields =
        WithProps.Find.byContainsOrdered<'TDoc> tableName criteria orderFields (fromDataSource ())

    /// Retrieve documents matching a JSON containment query (@>) ordered by the given fields in the document
    let ByContainsOrdered<'TDoc>(tableName, criteria: obj, orderFields) =
        WithProps.Find.ByContainsOrdered<'TDoc>(tableName, criteria, orderFields, fromDataSource ())

    /// Retrieve documents matching a JSON Path match query (@?)
    [<CompiledName "FSharpByJsonPath">]
    let byJsonPath<'TDoc> tableName jsonPath =
        WithProps.Find.byJsonPath<'TDoc> tableName jsonPath (fromDataSource ())

    /// Retrieve documents matching a JSON Path match query (@?)
    let ByJsonPath<'TDoc>(tableName, jsonPath) =
        WithProps.Find.ByJsonPath<'TDoc>(tableName, jsonPath, fromDataSource ())

    /// Retrieve documents matching a JSON Path match query (@?) ordered by the given fields in the document
    [<CompiledName "FSharpByJsonPathOrdered">]
    let byJsonPathOrdered<'TDoc> tableName jsonPath orderFields =
        WithProps.Find.byJsonPathOrdered<'TDoc> tableName jsonPath orderFields (fromDataSource ())

    /// Retrieve documents matching a JSON Path match query (@?) ordered by the given fields in the document
    let ByJsonPathOrdered<'TDoc>(tableName, jsonPath, orderFields) =
        WithProps.Find.ByJsonPathOrdered<'TDoc>(tableName, jsonPath, orderFields, fromDataSource ())

    /// Retrieve the first document matching a JSON field comparison query (->> =); returns None if not found
    [<CompiledName "FSharpFirstByFields">]
    let firstByFields<'TDoc> tableName howMatched fields =
        WithProps.Find.firstByFields<'TDoc> tableName howMatched fields (fromDataSource ())

    /// Retrieve the first document matching a JSON field comparison query (->> =); returns null if not found
    let FirstByFields<'TDoc when 'TDoc: null and 'TDoc: not struct>(tableName, howMatched, fields) =
        WithProps.Find.FirstByFields<'TDoc>(tableName, howMatched, fields, fromDataSource ())

    /// Retrieve the first document matching a JSON field comparison query (->> =) ordered by the given fields in the
    /// document; returns None if not found
    [<CompiledName "FSharpFirstByFieldsOrdered">]
    let firstByFieldsOrdered<'TDoc> tableName howMatched queryFields orderFields =
        WithProps.Find.firstByFieldsOrdered<'TDoc> tableName howMatched queryFields orderFields (fromDataSource ())

    /// Retrieve the first document matching a JSON field comparison query (->> =) ordered by the given fields in the
    /// document; returns null if not found
    let FirstByFieldsOrdered<'TDoc when 'TDoc: null and 'TDoc: not struct>(
            tableName, howMatched, queryFields, orderFields) =
        WithProps.Find.FirstByFieldsOrdered<'TDoc>(tableName, howMatched, queryFields, orderFields, fromDataSource ())

    /// Retrieve the first document matching a JSON containment query (@>); returns None if not found
    [<CompiledName "FSharpFirstByContains">]
    let firstByContains<'TDoc> tableName (criteria: obj) =
        WithProps.Find.firstByContains<'TDoc> tableName criteria (fromDataSource ())

    /// Retrieve the first document matching a JSON containment query (@>); returns null if not found
    let FirstByContains<'TDoc when 'TDoc: null and 'TDoc: not struct>(tableName, criteria: obj) =
        WithProps.Find.FirstByContains<'TDoc>(tableName, criteria, fromDataSource ())

    /// Retrieve the first document matching a JSON containment query (@>) ordered by the given fields in the document;
    /// returns None if not found
    [<CompiledName "FSharpFirstByContainsOrdered">]
    let firstByContainsOrdered<'TDoc> tableName (criteria: obj) orderFields =
        WithProps.Find.firstByContainsOrdered<'TDoc> tableName criteria orderFields (fromDataSource ())

    /// Retrieve the first document matching a JSON containment query (@>) ordered by the given fields in the document;
    /// returns null if not found
    let FirstByContainsOrdered<'TDoc when 'TDoc: null and 'TDoc: not struct>(tableName, criteria: obj, orderFields) =
        WithProps.Find.FirstByContainsOrdered<'TDoc>(tableName, criteria, orderFields, fromDataSource ())

    /// Retrieve the first document matching a JSON Path match query (@?); returns None if not found
    [<CompiledName "FSharpFirstByJsonPath">]
    let firstByJsonPath<'TDoc> tableName jsonPath =
        WithProps.Find.firstByJsonPath<'TDoc> tableName jsonPath (fromDataSource ())

    /// Retrieve the first document matching a JSON Path match query (@?); returns null if not found
    let FirstByJsonPath<'TDoc when 'TDoc: null and 'TDoc: not struct>(tableName, jsonPath) =
        WithProps.Find.FirstByJsonPath<'TDoc>(tableName, jsonPath, fromDataSource ())

    /// Retrieve the first document matching a JSON Path match query (@?) ordered by the given fields in the document;
    /// returns None if not found
    [<CompiledName "FSharpFirstByJsonPathOrdered">]
    let firstByJsonPathOrdered<'TDoc> tableName jsonPath orderFields =
        WithProps.Find.firstByJsonPathOrdered<'TDoc> tableName jsonPath orderFields (fromDataSource ())

    /// Retrieve the first document matching a JSON Path match query (@?) ordered by the given fields in the document;
    /// returns null if not found
    let FirstByJsonPathOrdered<'TDoc when 'TDoc: null and 'TDoc: not struct>(tableName, jsonPath, orderFields) =
        WithProps.Find.FirstByJsonPathOrdered<'TDoc>(tableName, jsonPath, orderFields, fromDataSource ())


/// Commands to update documents
[<RequireQualifiedAccess>]
module Update =

    /// Update an entire document by its ID
    [<CompiledName "ById">]
    let byId tableName (docId: 'TKey) (document: 'TDoc) =
        WithProps.Update.byId tableName docId document (fromDataSource ())

    /// Update an entire document by its ID, using the provided function to obtain the ID from the document
    [<CompiledName "FSharpFullFunc">]
    let byFunc tableName (idFunc: 'TDoc -> 'TKey) (document: 'TDoc) =
        WithProps.Update.byFunc tableName idFunc document (fromDataSource ())

    /// Update an entire document by its ID, using the provided function to obtain the ID from the document
    let ByFunc(tableName, idFunc: System.Func<'TDoc, 'TKey>, document: 'TDoc) =
        WithProps.Update.ByFunc(tableName, idFunc, document, fromDataSource ())


/// Commands to patch (partially update) documents
[<RequireQualifiedAccess>]
module Patch =

    /// Patch a document by its ID
    [<CompiledName "ById">]
    let byId tableName (docId: 'TKey) (patch: 'TPatch) =
        WithProps.Patch.byId tableName docId patch (fromDataSource ())

    /// Patch documents using a JSON field comparison query in the WHERE clause (->> =)
    [<CompiledName "ByFields">]
    let byFields tableName howMatched fields (patch: 'TPatch) =
        WithProps.Patch.byFields tableName howMatched fields patch (fromDataSource ())

    /// Patch documents using a JSON containment query in the WHERE clause (@>)
    [<CompiledName "ByContains">]
    let byContains tableName (criteria: 'TCriteria) (patch: 'TPatch) =
        WithProps.Patch.byContains tableName criteria patch (fromDataSource ())

    /// Patch documents using a JSON Path match query in the WHERE clause (@?)
    [<CompiledName "ByJsonPath">]
    let byJsonPath tableName jsonPath (patch: 'TPatch) =
        WithProps.Patch.byJsonPath tableName jsonPath patch (fromDataSource ())


/// Commands to remove fields from documents
[<RequireQualifiedAccess>]
module RemoveFields =

    /// Remove fields from a document by the document's ID
    [<CompiledName "ById">]
    let byId tableName (docId: 'TKey) fieldNames =
        WithProps.RemoveFields.byId tableName docId fieldNames (fromDataSource ())

    /// Remove fields from documents via a comparison on JSON fields in the document
    [<CompiledName "ByFields">]
    let byFields tableName howMatched fields fieldNames =
        WithProps.RemoveFields.byFields tableName howMatched fields fieldNames (fromDataSource ())

    /// Remove fields from documents via a JSON containment query (@>)
    [<CompiledName "ByContains">]
    let byContains tableName (criteria: 'TContains) fieldNames =
        WithProps.RemoveFields.byContains tableName criteria fieldNames (fromDataSource ())

    /// Remove fields from documents via a JSON Path match query (@?)
    [<CompiledName "ByJsonPath">]
    let byJsonPath tableName jsonPath fieldNames =
        WithProps.RemoveFields.byJsonPath tableName jsonPath fieldNames (fromDataSource ())


/// Commands to delete documents
[<RequireQualifiedAccess>]
module Delete =

    /// Delete a document by its ID
    [<CompiledName "ById">]
    let byId tableName (docId: 'TKey) =
        WithProps.Delete.byId tableName docId (fromDataSource ())

    /// Delete documents by matching a JSON field comparison query (->> =)
    [<CompiledName "ByFields">]
    let byFields tableName howMatched fields =
        WithProps.Delete.byFields tableName howMatched fields (fromDataSource ())

    /// Delete documents by matching a JSON containment query (@>)
    [<CompiledName "ByContains">]
    let byContains tableName (criteria: 'TContains) =
        WithProps.Delete.byContains tableName criteria (fromDataSource ())

    /// Delete documents by matching a JSON Path match query (@?)
    [<CompiledName "ByJsonPath">]
    let byJsonPath tableName path =
        WithProps.Delete.byJsonPath tableName path (fromDataSource ())
